// ArmControlMain.c
/* Sam Bobb,  Daniel Cornew,  Ryan Deeter
	Two degree of freedom arm
	High speed motor control for 2 motors
	Code version 3
	March 18, 2010
	
	Timer 1 - Motor 2 A Enc
	Timer 2 - Motor 2 B Enc
	Timer 3 - PWM Clock source
	Timer 4 - Motor 1 A Enc
	Timer 5 - Motor 1 B Enc
	
	Motor direction convention: increasing theta moves the arm up, decreasing moves it down. That actually means that the
	motors are flipped with respect to each other (because the shafts are facing each other). So Forward is clockwise on one
	motor and counterclockwise on the other.
	
	There are a lot of arrays with 2 elements. For example, int X[2]. In this case X[0] is for motor 1 and X[1] is motor 2.

*/

/** INCLUDES ***************************************************/
#include "HardwareProfile.h" //sets up the PIC board
#include "constants.h" // adds our defines and global vars
#include "MotorCommands.h" // functions for interfacing with the motor pins
#include "SerialFunctions.h" // functions for handling serial control

/** Main Function **********************************************/

int main(void)
{
	// Vars
	int	pbClk;
	int ref;
	signed int error[2];

	// --- Initiallization ---
	// Configure the proper PB frequency and the number of wait states
	pbClk = SYSTEMConfigPerformance(SYS_FREQ);
	// Allow vector interrupts
	INTEnableSystemMultiVectoredInt();
	mInitAllLEDs();	
	initPWMandIO();
	initEncoder();
	initUART2(pbClk);
	
	// --- Zero the arm ---
	// On startup, the user must move the arm through each of the optical
	// break sensors to establish an absolute position for the arm.
	
	// floats the motors so the user can freely move the arm
	motor_float();
	
	// these LEDs indicate the homing of arms
	// One goes off for each arm
	mLED_0_On();
	mLED_1_On();
	
	// After everything is initialized, wait for homing.
	// The user needs to move the arms 
	// First move arm 2
	while(  ReadADC10(1) < OPTO_BREAK_THREASHOLD ){}
	mLED_0_Off();
	// Now reset arm 2 counter
	WriteTimer1(0x00);
	WriteTimer2(0x00);
	currentTheta[1] = 0;
	lastT1 = 0;
	lastT2 = 0;
	// Now move arm 1
	while(  ReadADC10(0) < OPTO_BREAK_THREASHOLD ){}
	mLED_1_Off();
	// Now reset arm 1 counter
	WriteTimer4(0x00);
	WriteTimer5(0x00);
	currentTheta[0] = 0;
	lastT4 = 0;
	lastT5 = 0;
	
	// --- Main Loop ---
	while(1==1)
	{
		// We can put in here things that we want to happen 
		// while the arm is running. Control calculations will
		// interrupt these processes.
		if ((sendLog==TRUE) && (activeLogging==FALSE))
		{
			sendDataRS232();
			sendLog=FALSE;
			activeLogging=TRUE;
		}
	}
	// end of Main Loop
	// never get here
} //end main


/** Interrupt Handlers *****************************************/
// interrput code for the timer 3
// Using Timer 3 for both PWM clock period and our timed ISR
// Timer 3 is configured for 20 kHz
void __ISR( _TIMER_3_VECTOR, ipl7) T3Interrupt( void) //< = !!!!!
{	
	count_20khz++; // increment the counter
	// Watch out! Since we're doing these control calculations in the ISR, they have to take
	// less than 1/(20 kHz).
	// You can check how long your calculations take by monitoring the LOOP_TIME_PIN
	// check if 10 interupts have elapsed
	if(count_20khz >= CONTROL_DIV)
	{
		LOOP_TIME_PIN = TRUE;
		count_20khz = 0; // reset the count
		
		// update the motor position reading
		updateCurrentTheta();
		
		// Check our limits
		if ( (currentTheta[0] < LIMIT_ARM_1_DN) || (currentTheta[0] > LIMIT_ARM_1_UP) || (currentTheta[1] < LIMIT_ARM_2_DN) || (currentTheta[1] > LIMIT_ARM_2_UP) || ((currentTheta[1]-currentTheta[0]) < LIMIT_ARMS_OPEN) || ((currentTheta[1]-currentTheta[0]) > LIMIT_ARMS_CLOSED))
		{
			// Stop Everything!
			LIMIT_LED = 0; // LED on
			motor_break();
		}
		else{LIMIT_LED = 1;}
		
		if (runcontrol == TRUE)
		{	
			// update the control cycle
			// run the control update code
			// runs at PWM frequency / CONTROL_DIV
			// with current settings: 20 kHz / 10 = 2 kHz --> every .5 ms
			controlCount++; // increment the counter
		
			// update the current reference position
			updateReference();
			
			// Calculate the control output, update PWM outputs
			doControl();

			// Data Logging
			// check if 20 interupts have elapsed
			if((controlCount >= RECORD_EVERY) && (activeLogging==TRUE))
			{
				// log data
				refTheta1Log[recordIndex] = refTheta[0];
				refTheta2Log[recordIndex] = refTheta[1];
				currentTheta1Log[recordIndex] = currentTheta[0];
				currentTheta2Log[recordIndex] = currentTheta[1];
				recordIndex++;
				if (recordIndex > ((refPeriod/RECORD_EVERY)-1))
				{
					// we filled the buffer, so start over
					recordIndex = 0;
					
					if(sendLog == TRUE)
					{
						LOG_LED = 0; // led on
						// stop logging so we can send the log
						activeLogging = FALSE;
					}
				}
				controlCount = 0;
			}

			// increment globalIndex, set to zero at end of motion time
			globalIndex++;
			 if (globalIndex > refPeriod-1)
			 {
				// done with a path cycle
				 globalIndex = 0;
				 recordIndex = 0;
				 if (loop == FALSE)
				 {
					// stop running and hold in place
					runcontrol = FALSE;
					holdArm = TRUE;
				}
			 }
		}
		else if (holdArm == TRUE)
		{
			doHoldControl();  // hold at rest
			if(sendLog == TRUE)
			{
			// stop logging so we can send the log
				activeLogging = FALSE;
			}
		}
		else
		{
			if(sendLog == TRUE)
			{
			// stop logging so we can send the log
				activeLogging = FALSE;
			}
		
		}
		LOOP_TIME_PIN = FALSE;
	}
	
	// clear interrupt flag and exit
	mT3ClearIntFlag();
	
} // T3 Interrupt


// UART 2 interrupt handler
// it is set at priority level 2
void __ISR(_UART2_VECTOR, ipl2) IntUart2Handler(void)
{
	// Is this an RX interrupt?
	if(mU2RXGetIntFlag())
	{
		// Clear the RX interrupt Flag
	    mU2RXClearIntFlag();
		
		// Handle the new data
		process_incoming_data();
		 
	}

	// We don't care about TX interrupt
	if ( mU2TXGetIntFlag() )
	{
		mU2TXClearIntFlag();
	}
	
}


/** Other Functions ********************************************/
void initPWMandIO(void)
{
	//Setup digital outputs
	// Initialize values
	ENABLE_M1 = 0;
	DIRECTION_M1 = 0;
	ENABLE_M2 = 0;
	DIRECTION_M2 = 0;
	LOOP_TIME_PIN = 0;
	LIMIT_LED = 1;
	MEM_LED = 1;
	LOG_LED = 1;
	
	// set direction to outputs
	// motor pin
	TRISAbits.TRISA2 = 0;
	TRISAbits.TRISA3 = 0;
	TRISAbits.TRISA14 = 0;
	TRISAbits.TRISA15 = 0;
	
	TRISAbits.TRISA10 = 0; // limit LED pin
	TRISAbits.TRISA9 = 0; // mem led pin
	TRISEbits.TRISE9 = 0; // log led pin
	TRISDbits.TRISD10 = 0; // loop time pin
	
	// Setup the ADC for the optical indexers
	CloseADC10();	// ensure the ADC is off before setting the configuration

	// define setup parameters for OpenADC10
	// Turn module on | output in integer | trigger mode auto | enable  autosample
	#define PARAM1  ADC_MODULE_ON | ADC_FORMAT_INTG | ADC_CLK_AUTO | ADC_AUTO_SAMPLING_ON

	// define setup parameters for OpenADC10
	// ADC ref external    | disable offset test    | enable scan mode | perform 2 samples | use one buffer | use MUXA mode
    // note: to read X number of pins you must set ADC_SAMPLES_PER_INT_X
	#define PARAM2  ADC_VREF_AVDD_AVSS | ADC_OFFSET_CAL_DISABLE | ADC_SCAN_ON | ADC_SAMPLES_PER_INT_2 | ADC_ALT_BUF_OFF | ADC_ALT_INPUT_OFF

	// define setup parameters for OpenADC10
	// 				  use ADC internal clock | set sample time
	#define PARAM3  ADC_CONV_CLK_INTERNAL_RC | ADC_SAMPLE_TIME_15

	// define setup parameters for OpenADC10
	// set AN0 and AN1
	#define PARAM4	ENABLE_AN0_ANA | ENABLE_AN1_ANA

	// define setup parameters for OpenADC10
	// do not assign channels to scan
	#define PARAM5	SKIP_SCAN_AN2 | SKIP_SCAN_AN3 | SKIP_SCAN_AN4 | SKIP_SCAN_AN5 | SKIP_SCAN_AN6 | SKIP_SCAN_AN7 | SKIP_SCAN_AN8 | SKIP_SCAN_AN9 | SKIP_SCAN_AN10 | SKIP_SCAN_AN11 | SKIP_SCAN_AN12 | SKIP_SCAN_AN13 | SKIP_SCAN_AN14 | SKIP_SCAN_AN15

	// use ground as neg ref for A 
	SetChanADC10( ADC_CH0_NEG_SAMPLEA_NVREF); // use ground as the negative reference
	OpenADC10( PARAM1, PARAM2, PARAM3, PARAM4, PARAM5 ); // configure ADC using parameter define above

	EnableADC10(); // Enable the ADC

	while ( ! mAD1GetIntFlag() ) { } // wait for the first conversion to complete so there will be valid data in ADC result registers
	
	// Set up PWM and timers
	// Motor 1 PWM
	// init OC1 module, on pin D0 
	OpenOC1( OC_ON | OC_TIMER3_SRC | OC_PWM_FAULT_PIN_DISABLE, 0, 0);
	
	// Motor 2 PWM
	// init OC2 module, on pin D1
	OpenOC2( OC_ON | OC_TIMER3_SRC | OC_PWM_FAULT_PIN_DISABLE, 0, 0);
	
	// init Timer3 mode and period (PR2) // set for 20kHz
	// This generates the period for the PWM
	OpenTimer3( T3_ON | T3_PS_1_1 | T3_SOURCE_INT, 0x0F9F);
	
	// This lets Timer3 generate an interupt for our periodic code
	mT3SetIntPriority( 7); 	// set Timer3 Interrupt Priority   (highest)
	mT3ClearIntFlag(); 		// clear interrupt flag
	mT3IntEnable( 1);		// enable timer3 interrupts
	
	
}


void initEncoder(void)
{
	// init Timer1 and Timer2 mode and periods (Motor 2 encoder)
	OpenTimer1( T1_ON | T1_PS_1_1 | T1_SOURCE_EXT, 0xFFFF); 
	OpenTimer2( T2_ON | T2_PS_1_1 | T2_SOURCE_EXT, 0xFFFF);	
	
	// init Timer4 and Timer5 mode and periods (Moter 1 encoder)
	OpenTimer4( T4_ON | T4_PS_1_1 | T4_SOURCE_EXT, 0xFFFF); 
	OpenTimer5( T5_ON | T5_PS_1_1 | T5_SOURCE_EXT, 0xFFFF); 
} 

void initUART2(int pbClk)
{
	 // define setup Configuration 1 for OpenUARTx
 		// Module Enable 
 		// Work in IDLE mode 
 		// Communication through usual pins 
 		// Disable wake-up 
 		// Loop back disabled 
 		// Input to Capture module from ICx pin 
 		// no parity 8 bit 
 		// 1 stop bit 
 		// IRDA encoder and decoder disabled 
 		// CTS and RTS pins are disabled 
 		// UxRX idle state is '1' 
 		// 16x baud clock - normal speed
 	#define config1 	UART_EN | UART_IDLE_CON | UART_RX_TX | UART_DIS_WAKE | UART_DIS_LOOPBACK | UART_DIS_ABAUD | UART_NO_PAR_8BIT | UART_1STOPBIT | UART_IRDA_DIS | UART_DIS_BCLK_CTS_RTS| UART_NORMAL_RX | UART_BRGH_SIXTEEN
	 
	 // define setup Configuration 2 for OpenUARTx
	 	// IrDA encoded UxTX idle state is '0'
	 	// Enable UxRX pin
	 	// Enable UxTX pin
	 	// Interrupt on transfer of every character to TSR 
	 	// Interrupt on every char received
	 	// Disable 9-bit address detect
	 	// Rx Buffer Over run status bit clear
	 #define config2		UART_TX_PIN_LOW | UART_RX_ENABLE | UART_TX_ENABLE | UART_INT_TX | UART_INT_RX_CHAR | UART_ADR_DETECT_DIS | UART_RX_OVERRUN_CLEAR	
 
 	// Open UART2 with config1 and config2
 	OpenUART2( config1, config2, pbClk/16/DESIRED_BAUDRATE-1);	// calculate actual BAUD generate value.
 		
 	// Configure UART2 RX Interrupt with priority 2
 	ConfigIntUART2(UART_INT_PR2 | UART_RX_INT_EN);
}

void updateCurrentTheta()
{
	short countT1 = ReadTimer1();  // in your routine this must be done at least every 32000 encoder counts to avoid rollover ambiguity
	short countT2 = ReadTimer2(); 	
	short countT4 = ReadTimer4();
	short countT5 = ReadTimer5();
	
	// --- Motor 2 ---
	currentTheta[1] -= countT1 - lastT1; // add on the recent up-counts, since the last time

	if (countT1 < lastT1)
	{
		currentTheta[1] -= 65536; // count only increments, so if it got lower it must have rolled over
	}

	lastT1 = countT1;

	currentTheta[1] += countT2 - lastT2; // we're not worrying about rollover of the 32 bit currentTheta total

	if (countT2 < lastT2)
	{
		currentTheta[1] += 65536;
	}

	lastT2 = countT2;

	// --- Motor 1 ---
	currentTheta[0] += countT4 - lastT4; // add on the recent up-counts, since the last time

	if (countT4 < lastT4)
	{
		currentTheta[0] += 65536; // count only increments, so if it got lower it must have rolled over
	}

	lastT4 = countT4;

	currentTheta[0] -= countT5 - lastT5; // we're not worrying about rollover of the 32 bit currentTheta total

	if (countT5 < lastT5)
	{
		currentTheta[0] -= 65536;
	}

	lastT5 = countT5;	

	
 }


void updateReference()
{
	short int i = 0;
	char done = FALSE;
	
	// Does a linear interpolation between path points. So we can tell the arm to be at 0inches,0inches at 1 second and at 20inches,20inches
	// 5 seconds later and it does a ramp function, not a step function.
	// formula is:
	// y = y0 + (x - x0)*(y1 - y0)/(x1 - x0)
	// The loop finds which two path time points the global time index is between and then interpolates between those path points.
	while(!done)
	{
		if(globalIndex < pathTime[i])
		{
			refTheta[0] = pathTheta1[i-1] + (globalIndex - pathTime[i-1]) * (pathTheta1[i] - pathTheta1[i-1]) / (pathTime[i] - pathTime[i-1]);
			refTheta[1] = pathTheta2[i-1] + (globalIndex - pathTime[i-1]) * (pathTheta2[i] - pathTheta2[i-1]) / (pathTime[i] - pathTime[i-1]);		
			done = TRUE;
		}
		else
		{
			i++;
		}

	}

}


void doControl()
{
	signed int error[2];
	signed int dError[2];
	signed int signedPWM[2];

	// calculate error
	// Motor 1
	error[0] = refTheta[0] - currentTheta[0];
	// Motor 2
	error[1] = refTheta[1] - currentTheta[1];
	
	// add to the error buffer
	errorBuffer[0][errorIndex] = error[0];
	errorBuffer[1][errorIndex] = error[1];
	
	// calculate derivative of error (one step back)
	//dError[0]  = error[0] - lastError[0];
	//dError[1]  = error[1] - lastError[1];
	
	// calculate derivative of error (better way) -- subtracts the oldest error stored from the current error
	if (errorIndex == ERROR_DERIV_DELAY-1)
	{
		// we're at the end of the array, so the oldest one is at the beginning (it's a circular buffer)
		dError[0] = error[0] - errorBuffer[0][0];
		dError[1] = error[1] - errorBuffer[1][0];
	}
	else
	{
		// otherwise the oldest sample is the one right in front of the current
		dError[0] = error[0] - errorBuffer[0][errorIndex+1];
		dError[1] = error[1] - errorBuffer[1][errorIndex+1];
	}
	
	// update last error
	//lastError[0] = error[0];
	//lastError[1] = error[1];
	
	// calculate integral of error
	integralError[0] += error[0];
	integralError[1] += error[1];
	
	// anti-windup
	if (integralError[0] > MAX_INTEGRAL)
	{
		integralError[0] = MAX_INTEGRAL;
	}
	else if (integralError[0] < -MAX_INTEGRAL)
	{
		integralError[0] = -MAX_INTEGRAL;
	}
	
	if (integralError[1] > MAX_INTEGRAL)
	{
		integralError[1] = MAX_INTEGRAL;
	}
	else if (integralError[1] < -MAX_INTEGRAL)
	{
		integralError[1] = -MAX_INTEGRAL;
	}
	

	// increment errorIndex
	errorIndex++;
	if (errorIndex == ERROR_DERIV_DELAY)
	{
		errorIndex = 0;
	}
	
	
	// control calculation
	signedPWM[0] = error[0]*KPnum[0]/KPden[0] + dError[0]*KDnum[0]/KDden[0] + integralError[0]*KInum[0]/KIden[0];
	signedPWM[1] = error[1]*KPnum[1]/KPden[1] + dError[0]*KDnum[1]/KDden[1] + integralError[0]*KInum[1]/KIden[1];
	
	// this is the unimplemented feed forward algorithm. needs more work.
	//signedPWM[0] = error[0]*KPnum[0]/KPden[0] + dError[0]*KDnum[0]/KDden[0] + integralError[0]*KInum[0]/KIden[0] + (refSlope[0]/Kslope[0]);
	//signedPWM[1] = error[1]*KPnum[1]/KPden[1] + dError[0]*KDnum[1]/KDden[1] + integralError[0]*KInum[1]/KIden[1] + (refSlope[1]/Kslope[1]);
	
	
	// check for saturation
	if (signedPWM[0] > MAX_RESOLUTION)
	{
		signedPWM[0] = MAX_RESOLUTION;
	}
	else if (signedPWM[0] < -MAX_RESOLUTION)
	{
		signedPWM[0] = -MAX_RESOLUTION;
	}
	
	if (signedPWM[1] > MAX_RESOLUTION)
	{
		signedPWM[1] = MAX_RESOLUTION;
	}
	else if (signedPWM[1] < -MAX_RESOLUTION)
	{
		signedPWM[1] = -MAX_RESOLUTION;
	}

	// send to the motors
	setPWMandDirection(signedPWM[0] , signedPWM[1]);
	
	// logging of the error
	// if you want to log error, you need to do it here.

	// check if 20 interupts have elapsed
	/*
	if(controlCount >= RECORD_EVERY)
	{
		// log data
		
		errorLog[recordIndex][0] = error[0];
		errorLog[recordIndex][1] = error[1];
		PWMLog[recordIndex][0] = signedPWM[0];
		PWMLog[recordIndex][1] = signedPWM[1];

		dErrorLog[recordIndex] = dError[0];
		iErrorLog[recordIndex] = integralError[0];
		
	}
	*/

}



void doHoldControl()
{
	signed int error[2];
	signed int dError[2];
	signed int signedPWM[2];
	


	// calculate error
	// Motor 1
	error[0] = (holdTheta[0] - currentTheta[0]);
	// Motor 2
	error[1] = (holdTheta[1] - currentTheta[1]);
	
	
	errorBuffer[0][errorIndex] = error[0];
	errorBuffer[1][errorIndex] = error[1];
	
	// calculate derivative of error (one step back)
	//dError[0]  = error[0] - lastError[0];
	//dError[1]  = error[1] - lastError[1];
	
	// calculate derivative of error (better way) -- subtracts the oldest error stored from the current error
	if (errorIndex == ERROR_DERIV_DELAY-1)
	{
		dError[0] = error[0] - errorBuffer[0][0];
		dError[1] = error[1] - errorBuffer[1][0];
	}
	else
	{
		dError[0] = error[0] - errorBuffer[0][errorIndex+1];
		dError[1] = error[1] - errorBuffer[1][errorIndex+1];
	}
	
	// update last error
	//lastError[0] = error[0];
	//lastError[1] = error[1];
	
	// calculate integral of error
	integralError[0] += error[0];
	integralError[1] += error[1];
	
	// anti-windup
	if (integralError[0] > MAX_INTEGRAL)
	{
		integralError[0] = MAX_INTEGRAL;
	}
	else if (integralError[0] < -MAX_INTEGRAL)
	{
		integralError[0] = -MAX_INTEGRAL;
	}
	
	if (integralError[1] > MAX_INTEGRAL)
	{
		integralError[1] = MAX_INTEGRAL;
	}
	else if (integralError[1] < -MAX_INTEGRAL)
	{
		integralError[1] = -MAX_INTEGRAL;
	}
	

	// increment errorIndex
	errorIndex++;
	if (errorIndex == ERROR_DERIV_DELAY)
	{
		errorIndex = 0;
	}
	
	
	// control calculation
	
	signedPWM[0] = error[0]*HOLD_KPnum/HOLD_KPden + dError[0]*HOLD_KDnum/HOLD_KDden + integralError[0]*HOLD_KInum/HOLD_KIden;
	signedPWM[1] = error[1]*HOLD_KPnum/HOLD_KPden + dError[0]*HOLD_KDnum/HOLD_KDden + integralError[0]*HOLD_KInum/HOLD_KIden;
	
	
	// check for saturation
	if (signedPWM[0] > MAX_RESOLUTION)
	{
		signedPWM[0] = MAX_RESOLUTION;
	}
	else if (signedPWM[0] < -MAX_RESOLUTION)
	{
		signedPWM[0] = -MAX_RESOLUTION;
	}
	
	if (signedPWM[1] > MAX_RESOLUTION)
	{
		signedPWM[1] = MAX_RESOLUTION;
	}
	else if (signedPWM[1] < -MAX_RESOLUTION)
	{
		signedPWM[1] = -MAX_RESOLUTION;
	}

	setPWMandDirection(signedPWM[0] , signedPWM[1]);
	

}





