// main.h
/* Sam Bobb,  Daniel Cornew,  Ryan Deeter
	Two degree of freedom arm
	High speed motor control for 2 motors
	Code version 3
	March 18, 2010
	

*/

#ifndef __MAIN_H
#define __MAIN_H

/** Constants **************************************************/

#define TRUE 		1
#define FALSE		0

// Motor Pins
#define ENABLE_M1				LATAbits.LATA2
#define DIRECTION_M1		LATAbits.LATA3 
#define ENABLE_M2				LATAbits.LATA14
#define DIRECTION_M2		LATAbits.LATA15

// Debug Pins
#define LOOP_TIME_PIN		LATDbits.LATD10
#define LIMIT_LED				LATAbits.LATA10
#define MEM_LED					LATAbits.LATA9
#define LOG_LED					LATEbits.LATE9

// Motor Directions
#define FORWARD			0
#define REVERSE				1

// parameters
#define MAX_RESOLUTION			0x0F9F		// Proportional to period of PWM
#define DESIRED_BAUDRATE    	(19200)      // The desired BaudRate 
#define CONTROL_DIV			10		// clock (timer 3 or 20kHz) divided by CONTROL_DIV gives the control update freq
#define RECORD_EVERY			20 		// how often do we log data
#define MAX_INTEGRAL			1000		// anti windup integral term
#define OPTO_BREAK_THREASHOLD 	100	// threshold for detected a break in the opto sensor
#define ERROR_DERIV_DELAY				10 // how many control calculations back should we look to calculate the derivative


// Serial commands
// Define the same ones in MATLAB
#define C_HOLD						0 // Hold arm in home position
#define C_FLOAT						1 // Stop and float motors
#define C_RUN						2 // Run
#define C_CONTROL_CONFIG	3 // Send new control parameters
#define C_LOG_CONFIG			4 // Send new logging parameters
#define C_SET_PATH				5 // Send new path points (format: num_points, time, theta1, theta2, time, theta1, theta2, ... )
#define C_SEND_LOG				6 // Orders the PIC to send the logged data
#define C_LOOP						7 // 0 = run once, 1 = loop the path
#define C_STARTSTOP				8 // start and stop commands
#define C_GOTO						9  // tells the arm to goto a certain X Y
#define C_PAUSE						10 // Stop and hold at current location. Sets the reference to the current XY and keeps the controller running.

// Configuration/programming constants
#define NUM_CONTROL_PARAMS	12 // kpnum, kpden, kinum, kiden, kdnum, kdden  <- for each motor
#define NUM_LOG_PARAMS		1 // record frequency, ...
#define NUM_PATH_PARAMS		2 // polynomial or array, polynomial power or number of data points 

// Arm Limits
// These are the encoder counts that provide limits for the arms so it doesn't break itself
// We check for each arm going too far up and down and we check for the arms too far open and too far closed.
#define LIMIT_ARM_1_DN			-10400
#define LIMIT_ARM_1_UP			8500
#define LIMIT_ARM_2_DN			-9500
#define LIMIT_ARM_2_UP			9400
#define LIMIT_ARMS_OPEN		1500
#define LIMIT_ARMS_CLOSED	19000

// Hold Parameters
// These are some gentel control parameters that we use for the PAUSE function and the GOTO function
// We don't want to use what could be crazy control params set by the user for these functions
#define M1_HOLD_THETA		-4875 // 45 degress
#define M2_HOLD_THETA		4875
#define HOLD_KPnum			15
#define HOLD_KPden			1
#define HOLD_KDnum			10
#define HOLD_KDden			1
#define HOLD_KInum			1
#define HOLD_KIden			1000

#define GOTO_DURATION_TICKS	10000 // 5 seconds


/** Function Declarations **************************************/
void initPWMandIO();

void initEncoder(void);

void updateCurrentTheta(void);

void updateReference(void);

void doControl(void);

void doHoldControl(void);

void setPWMandDirection(signed int M1signedPWM, signed int M2signedPWM);

void initUART2(int pbClk);


/** Global Variables *******************************************/

short lastT1 = 0, lastT2 = 0, lastT4 = 0, lastT5 = 0; // 16 bit number, holds previous timer values

char count_20khz = 0; // counts 20kHz interupts
char controlPeriodFlag = FALSE; // set to TRUE when we want to run control update

signed int refTheta[2] = {0, 0}; // holds the reference theta

// For logging these values.
// if you want to do this in a general way, you'll need to add mallocs for these
// in the serial functions.
//signed short int refThetaLog[NUM_DATA_POINTS][2];
//signed short int currentThetaLog[NUM_DATA_POINTS][2];
//signed short int errorLog[NUM_DATA_POINTS][2];
//signed short int dErrorLog[NUM_DATA_POINTS];
//signed short int iErrorLog[NUM_DATA_POINTS];
//signed short int PWMLog[NUM_DATA_POINTS][2];

//signed int lastError[2] = {0, 0}; // holds last error for calculating the derivative (we're doing this with an array now)
signed int integralError[2] = {0, 0};

short unsigned int recordIndex = 0;

char RS232_Out_Buffer[32];

signed int currentTheta[2] = {0, 0}; 	// set encoder value initially to zero, it can go + or -  (32 bit number)

int globalIndex = 0; // counts global time index

char controlCount = 0;  // counts how many control calcs we've done so we know when to record a log point. reset when log is made.

// control parameters
// each is defined with a numerator and denominator
// the slope parameter is an unimplemented feed forward controller
signed short int KPnum[2] = {0,1}; // proportional gain
signed short int KPden[2] = {1,1};
signed short int KDnum[2] = {0,0}; // derivative gain
signed short int KDden[2] = {1,1};
signed short int KInum[2] = {0,0}; // integral gain
signed short int KIden[2] = {1,1};
signed short int Kslope[2] = {1,1};

// holds past errors, used for derivative calculation
signed int errorBuffer[2][ERROR_DERIV_DELAY] = {0};

// indexes the errorBuffer
short int errorIndex = 0;

// the reference theta to hold the arm at.
// used for the goto and pause functions.
signed int holdTheta[2] = {0, 0};
	
// the period of our motion path
// in ticks (0.5 ms in this case)
int refPeriod = 0;		

// The motion path is defined as sets like: (time, theta1, theta2)
// That gets stored in three vectors (we'll malloc them later to adjust the length as needed):
signed short int * pathTime;
signed short int * pathTheta1;
signed short int * pathTheta2;

// The log consists of the reference theta and actual theta for each motor.
// That data is stored in the following vectors which we'll set the size of dynamically with malloc
signed short int * refTheta1Log;
signed short int * currentTheta1Log;
signed short int * refTheta2Log;
signed short int * currentTheta2Log;

// Flags
char loop = FALSE;
char sendLog = FALSE;
char activeLogging = TRUE;
char runcontrol = FALSE;
char holdArm = FALSE;
char runcontrol2 = 0;
int recordData = FALSE;

#endif
